"""Azure blob storage helper functions"""

import os
from typing import Optional, Tuple

from azure.core.exceptions import AzureError
from azure.storage.blob import BlobServiceClient


def _validate_access_info(
    account_name: Optional[str], blob_container_name: Optional[str], sas_credential: Optional[str]
) -> bool:
    """Validates the retrieved access info"""

    return all([account_name, blob_container_name, sas_credential])


def _get_access_info() -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """Retreives access information from the environment variables"""

    account_name = os.getenv("AZURE_ACCOUNT_NAME")
    blob_container_name = os.getenv("AZURE_BLOB_CONTAINER_NAME")
    sas_credential = os.getenv("AZURE_SAS_CREDENTIAL")

    return account_name, blob_container_name, sas_credential


def write_file_to_azure_blob_storage(
    local_file_path: str, blob_name: Optional[str] = None
) -> bool:
    """Upload file to Azure Blob Storage.

    Args:
      local_file_path(str): Path of the file to upload.
      blob_name(str, optional): Name of the blob to be uploaded, defaults to None

    Returns:
      bool: True or False depending on the success of the upload.
    """

    account_name, blob_container_name, sas_credential = _get_access_info()

    if _validate_access_info(account_name, blob_container_name, sas_credential):
        try:
            if blob_name is None:
                blob_name = os.path.basename(local_file_path)
            account_url = f"https://{account_name}.blob.core.windows.net"

            # Upload file
            blob_service_client = BlobServiceClient(
                account_url=account_url, credential=sas_credential
            )

            if blob_container_name is not None:
                blob_client = blob_service_client.get_blob_client(blob_container_name, blob_name)
                with open(local_file_path, "rb") as local_file:
                    blob_client.upload_blob(local_file)

                return True

            return False

        except AzureError as error:
            print(f"Error while uploading file to Azure: {error}")
            return False

    else:
        print("Please enter valid access information to Azure blob storage in .env file")
        return False


def upload_directory_to_azure_blob_storage(local_directory_path: str) -> bool:
    """Upload a directory and its contents to Azure Blob Storage.

    Args:
      local_directory_path(str): The local directory path to upload.

    Returns:
      bool: True or False depending on the success of the upload.
    """

    account_name, blob_container_name, sas_credential = _get_access_info()
    directory_name = os.path.basename(local_directory_path)
    if _validate_access_info(account_name, blob_container_name, sas_credential):
        try:
            for root, _, files in os.walk(local_directory_path):
                for file in files:
                    local_file_path = os.path.join(root, file)

                    relative_path = os.path.relpath(local_file_path, local_directory_path)
                    blob_name = os.path.join(directory_name, relative_path).replace(os.sep, "/")
                    write_file_to_azure_blob_storage(local_file_path, blob_name)

            return True

        except AzureError as error:
            print(f"Error while uploading directory to Azure Blob Storage: {error}")
            return False

    else:
        print("Please enter valid access information to Azure blob storage in .env file")
        return False


def get_azure_blob_url_path(local_path: str) -> str:
    """Get the azure blob path of the uploaded element (file or folder)

    Args:
      local_path(str): The local path of the uploaded folder/file

    Returns:
      str: The azure blob url of the uploaded folder/file
    """

    azure_blob_key = os.path.basename(local_path)
    account_name, blob_container_name, _ = _get_access_info()

    blob_url = (
        f"https://{account_name}.blob.core.windows.net/{blob_container_name}/{azure_blob_key}"
    )
    return blob_url



"""Context command"""

from pathlib import Path

import click

from byoa.config.context import (
    DEFAULT_CONFIG_DIR,
    ENDPOINT_KEY,
    REGISTRY_KEY,
    append_context,
    get_context_value,
)


@click.command("context", short_help="sets the BYOA context")
@click.option("--profile", default="default", help="Named context")
@click.option(
    "--config-dir",
    default=DEFAULT_CONFIG_DIR,
    show_default=True,
    type=Path,
    help="Context file path",
)
def context(profile: str, config_dir: Path):
    """Sets the BYOA context for deployement"""

    default_registry = get_context_value(REGISTRY_KEY, profile, config_dir=config_dir)
    default_endpoint = get_context_value(ENDPOINT_KEY, profile, config_dir=config_dir)

    registry = click.prompt("Container registry name", default_registry)
    endpoint = click.prompt("BYOA API endpoint", default_endpoint)

    append_context(registry, endpoint, profile, config_dir)



"""BYOA exceptions"""

from byoa import MANIFEST_FILE


class ByoaException(Exception):
    """Base BYOA error"""


class NoManifestFoundException(ByoaException):
    """Raised when the byoa manifest file is missing"""

    def __init__(self):
        super().__init__(
            f"No '{MANIFEST_FILE}' file found, make sure that you are inside a BYOA project."
        )


class UnknownKeyException(ByoaException):
    """Raised when the configuration key is not handled

    Args:
        key (str): The unknown configuration key
    """

    def __init__(self, key: str):
        super().__init__(f"Unknown configuration key '{key}'.")


class MissingContextValue(ByoaException):
    """Raised when an context value is missing"""

    def __init__(self, key: str):
        super().__init__(f"Missing context value for '{key}'")


class APIError(ByoaException):
    """Raised when an error occures communicating to the API"""


class ImagePushError(ByoaException):
    """Raised when an image push has failed"""

    def __init__(self, reason: str):
        super().__init__(f"Could not push processor image: '{reason}'")



"""Logging functions"""

import logging
from typing import Optional


# pylint: disable=locally-disabled, too-few-public-methods
class LogManager:
    """
    Class to create customized logger instances.

    Attributes:
        logger (logging.Logger): Instance of logger configured for LogManager.

    Methods:
        _configure_logger(handlers=None, formatter=None): Private method to configure the logger.
    """

    _shared_log_manager: Optional["LogManager"] = None

    def __init__(self, handlers=None, formatter=None):
        """
        Initializes an instance of Logging with one or more log handlers.

        Args:
            handlers (list, optional): List of log handlers. StreamHandler by default.
            formatter (logging.Formatter, optional): Logger's formatter. None by default.

        Returns:
            None
        """
        self.logger = self._configure_logger(handlers, formatter)

        if not LogManager._shared_log_manager:
            LogManager._shared_log_manager = self

    @classmethod
    def get_instance(cls, handlers=None, formatter=None) -> logging.Logger:
        """_summary_

        Args:
            handlers (list, optional): List of log handlers. StreamHandler by default.
            formatter (logging.Formatter, optional): Logger's formatter. None by default.

        Returns:
            logging.Logger: Configured logger instance.
        """
        if not cls._shared_log_manager:
            cls._shared_log_manager = cls(handlers, formatter)
        return cls._shared_log_manager.logger

    def _configure_logger(self, handlers=None, formatter=None):
        """
        Configures the logger with specified log handlers.

        Args:
            handlers (list, optional): List of log handlers. StreamHandler by default.
            formatter (logging.Formatter, optional): Logger's formatter. None by default.

        Returns:
            logging.Logger: Configured logger instance.
        """
        logger = logging.getLogger(self.__class__.__name__)
        logger.setLevel(logging.INFO)

        formatter = formatter or logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        if handlers is None:
            handlers = [logging.StreamHandler()]  # By default, use a console handler

        for handler in handlers:
            handler.setLevel(logging.DEBUG)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

        return logger



"""Deploy command"""

import click
import docker
import docker.errors
import requests

from byoa.config.context import ENDPOINT_KEY, REGISTRY_KEY, get_context_value
from byoa.config.manifest import Manifest, check_manifest
from byoa.exceptions import APIError, ImagePushError, MissingContextValue
from byoa.importer import get_variable


def _login_to_docker(docker_client: docker.DockerClient, registry: str):
    username = click.prompt("Username")
    password = click.prompt("Password", hide_input=True)

    click.echo("Login to registry... ", nl=False)
    result_login = docker_client.login(username=username, password=password, registry=registry)
    click.echo(result_login["Status"])


def _save_processor(processor_id: str, data: dict, endpoint: str):
    try:
        click.echo(f"Saving processor {processor_id}... ", nl=False)
        response = requests.put(f"{endpoint}/process_graphs/{processor_id}", json=data, timeout=60)
        response.raise_for_status()
        click.echo("Done!")

    except requests.HTTPError as e:
        raise APIError(f'Cannot save processor {processor_id}: "{e}"') from e


def _delete_processor(processor_id: str, endpoint: str):
    try:
        response = requests.delete(f"{endpoint}/process_graphs/{processor_id}", timeout=60)
        response.raise_for_status()

    except requests.HTTPError as e:
        raise APIError(f'Cannot delete processor {processor_id}: "{e}"') from e


@click.command("deploy", short_help="deploys the processor")
@click.option(
    "--registry",
    help="Container registry name",
    envvar="BYOA_CONTAINER_REGISTRY",
    show_envvar=True,
    default=lambda: get_context_value(REGISTRY_KEY),
)
@click.option(
    "--endpoint",
    help="BYOA API endpoint",
    envvar="BYOA_ENDPOINT",
    show_envvar=True,
    default=lambda: get_context_value(ENDPOINT_KEY),
)
@click.option("--profile", help="Profile to use for environment values")
@check_manifest
def deploy(profile: str, registry: str, endpoint: str):
    """Saves and deploys the processor"""

    if profile:
        _registry = get_context_value(REGISTRY_KEY, profile, registry)
        _endpoint = get_context_value(ENDPOINT_KEY, profile, endpoint)

        if _endpoint is None:
            raise MissingContextValue(ENDPOINT_KEY)
        if _registry is None:
            raise MissingContextValue(REGISTRY_KEY)

        endpoint = _endpoint
        registry = _registry

    manifest = Manifest()

    with open("VERSION", "r", encoding="utf-8") as f:
        version = f.read()

    output_files = get_variable("schemas.output_schema", "OUTPUT_FILES")
    links = [
        {"rel": "output_file", "href": file.local_path, "title": file.name, "type": file.type}
        for file in output_files
    ]

    links.append(
        {
            "rel": "image",
            "href": f"{_registry}/byoa/processor/{manifest.slug}:v{version}",
        }
    )

    data = {
        "id": manifest.slug,
        "summary": manifest.summary,
        "description": manifest.description,
        "categories": ["byoa"],
        "parameters": [
            {"name": "input_data", "description": "Input data", "schema": {"type": "object"}}
        ],
        "returns": {"description": "Output data", "schema": {"type": "object"}},
        "process_graph": {},
        "links": links,
    }

    _save_processor(manifest.slug, data, endpoint)

    docker_client = docker.from_env()

    click.echo("Building image... ", nl=False)
    docker_client.images.build(
        path="./", tag=f"{_registry}/byoa/processor/{manifest.slug}:v{version}"
    )
    click.echo("DONE !")

    if click.confirm("Do you need to login to your container registry ?", True):
        _login_to_docker(docker_client, registry)

    try:
        click.echo("Pushing the image... ", nl=False)
        docker_client.images.push(f"{_registry}/byoa/processor/{manifest.slug}", tag=f"v{version}")
        click.echo("DONE !")
    except docker.errors.APIError as e:
        click.echo("Reverting...")
        _delete_processor(manifest.slug, endpoint)
        raise ImagePushError(e.response.text) from e



"""Open telemetry Trace manager  """
from typing import Any, List, Optional

from opentelemetry import trace
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor


class TraceManager:
    """
    Class to handle OpenTelemetry Tracer instances with a singleton pattern.

    Attributes:
        _instance (TraceManager): Singleton instance of TraceManager.
        tracer (opentelemetry.trace.Tracer): OpenTelemetry tracer instance.
        exporters (list): List of OpenTelemetry exporters.

    Methods:
        _initialize_tracer(): Initializes the OpenTelemetry Tracer.
        get_instance(): Retrieves the singleton instance of TraceManager.
        add_exporter(exporter): Adds an OpenTelemetry exporter to the TraceManager.
        remove_exporter(exporter): Removes an OpenTelemetry exporter from the TraceManager.
        configure_exporters(): Configures the exporters for the TraceManager.
    """

    _instance: Optional["TraceManager"] = None
    exporters: List[Any]
    tracer: Any

    def __new__(cls) -> "TraceManager":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.tracer = cls._instance._initialize_tracer()
            cls._instance.exporters = []
        return cls._instance

    def _initialize_tracer(self):
        # Configure OpenTelemetry Tracer Provider
        tracer_provider = TracerProvider(resource=Resource.create({SERVICE_NAME: "processor"}))
        trace.set_tracer_provider(tracer_provider)

        # Create a Console exporter
        span_console_processor = SimpleSpanProcessor(ConsoleSpanExporter())

        tracer_provider.add_span_processor(span_console_processor)

        return trace.get_tracer(__name__)

    @classmethod
    def get_instance(cls):
        """
        Retrieves the singleton instance of TraceManager
        """
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    def add_exporter(self, exporter):
        """
        Adds an OpenTelemetry exporter to the TraceManager.

        Args:
            exporter: OpenTelemetry exporter to be added.
        """
        if exporter not in self.exporters:
            self.exporters.append(exporter)

    def remove_exporter(self, exporter):
        """
        Removes an OpenTelemetry exporter from the TraceManager.

        Args:
            exporter: OpenTelemetry exporter to be removed.
        """
        if exporter in self.exporters:
            self.exporters.remove(exporter)



""""Import module"""

import importlib


class ModelImportError(Exception):
    """Exception raised when an error occures when importing a model class"""


def get_variable(module_name: str, variable_name: str):
    """Returns a variable from a module

    Args:
        module_name (str)
        variable_name (str)
    """

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        raise ModelImportError(f"Could not find module {module_name}") from e

    return getattr(module, variable_name)



"""Init command"""

from pathlib import Path

import click
from cookiecutter.main import cookiecutter
from slugify import slugify

from byoa.config.manifest import Manifest


def _get_project_slug():
    return slugify(click.get_current_context().params["name"], separator="_")


def _get_project_repo():
    return slugify(click.get_current_context().params["name"])


@click.command("init", short_help="creates a new processor project")
@click.argument(
    "directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path, writable=True),
    default=".",
)
@click.option("--advanced", "-a", is_flag=True, help="Creates an advanced project structure")
@click.option("--name", prompt=True, required=True, help="Name of the processor")
@click.option(
    "--slug",
    prompt=True,
    default=_get_project_slug,
    help="Human-readable, unique identifier of the processor",
)
@click.option(
    "--repository",
    prompt=True,
    default=_get_project_repo,
    help="The project's repository name",
)
@click.option(
    "--summary", prompt=True, default="", help="A short sentence describing the processor"
)
@click.option("--description", prompt=True, default="", help="Description of the processor")
@click.option("--author", prompt=True, default="", help="Author(s) of the project")
def init_processor(
    directory: Path,
    advanced: bool,
    name: str,
    slug: str,
    repository: str,
    summary: str,
    description: str,
    author: str,
):
    # pylint: disable=too-many-arguments
    """
    Creates a new BYOA project in the current directory or in DIRECTORY if specified.
    """

    cookiecutter(
        "https://github.com/GEOSYS/analytic-processor-template",
        no_input=True,
        extra_context={
            "project_name": name,
            "project_repo": repository,
            "project_slug": slug,
            "author_name": author,
            "description": description,
            "mode": "advanced" if advanced else "basic",
        },
        output_dir=directory,
        checkout="new-version-abj",
    )

    click.echo("Create byoa manifest file")
    manifest = Manifest(from_file=False)
    manifest.set("name", name)
    manifest.set("slug", slug)
    manifest.set("summary", summary)
    manifest.set("description", description)
    manifest.set("author", author)
    manifest.save(directory=directory.joinpath(repository))
